Option Strict On
Option Explicit On

Imports System.Runtime.InteropServices

' Class that holds methods for fitting either an ARIMA or Multi-Input time series model
' Forecasts can be obtained from these models. There are multiple ways of structuring
' this, so this is just one way and not necessarily the "best" way. 
Class G13B

    ' Subroutine for a model estimation for a univariate ARIMA 
    ' No forecasting and default options
    Public Overloads Sub UnivariateARIMA(ByVal arimav As Nag_ArimaOrder, _
     ByRef para() As Double, ByVal y() As Double, ByRef sd() As Double, _
     ByRef rss As Double, ByRef objf As Double, ByRef df As Double)
        Dim nfv As Integer
        Dim rmsy As Double
        Dim fva(0) As Double, fsd(0) As Double
        nfv = 0

        ' An alternative way of using the default options is to supply a
        ' null pointer to the options argument, i.e. a value of 0, when calling G13BE and G13BJ.
        '  However in order to do this multiple declare statements would be required and the 
        ' "options" argument would need to be declared as ByVal and an Integer

        ' Initialise the options structure
        Dim options As New Nag_G13_Opt()
        Call g13bxc(options)

        ' Turn off all printing etc
        options.list = Nag_Boolean.Nag_FALSE
        options.print_level = Nag_PrintType.Nag_NoPrint

        Call UnivariateARIMA(arimav, para, y, sd, rss, objf, df, nfv, fva, fsd, rmsy, options)

        ' Free the options structure. This is required as G13BE and G13BJ may allocate memory 
        ' to various elements in the structure.
        Call g13xzc(options)
    End Sub

    ' Subroutine for a model estimation for a univariate ARIMA 
    ' No forecasting and options supplied explicitly
    Public Overloads Sub UnivariateARIMA(ByVal arimav As Nag_ArimaOrder, _
    ByRef para() As Double, ByVal y() As Double, ByRef sd() As Double, _
    ByRef rss As Double, ByRef objf As Double, ByRef df As Double, _
    ByRef options As Nag_G13_Opt)
        Dim nfv As Integer
        Dim rmsy As Double
        Dim fva(0) As Double, fsd(0) As Double
        nfv = 0
        Call UnivariateARIMA(arimav, para, y, sd, rss, objf, df, nfv, fva, fsd, rmsy, options)
    End Sub

    ' Subroutine for a model estimation for a univariate ARIMA
    ' With forecasting and default options
    Public Overloads Sub UnivariateARIMA(ByVal arimav As Nag_ArimaOrder, _
     ByRef para() As Double, ByVal y() As Double, ByRef sd() As Double, _
     ByRef rss As Double, ByRef objf As Double, ByRef df As Double, _
     ByVal nfv As Integer, ByRef fva() As Double, ByRef fsd() As Double, _
     ByRef rmsy As Double)

        ' Initialise the options structure
        Dim options As New Nag_G13_Opt()
        Call g13bxc(options)

        ' Turn off all printing etc
        options.list = Nag_Boolean.Nag_FALSE
        options.print_level = Nag_PrintType.Nag_NoPrint

        Call UnivariateARIMA(arimav, para, y, sd, rss, objf, df, nfv, fva, fsd, rmsy, options)

        ' Free the options structure. This is required as G13BE and G13BJ may allocate memory 
        ' to various elements in the structure.
        Call g13xzc(options)
    End Sub

    ' Subroutine for a model estimation for a univariate ARIMA
    ' With forecasting and options supplied explicitly
    Public Overloads Sub UnivariateARIMA(ByVal arimav As Nag_ArimaOrder, _
     ByRef para() As Double, ByVal y() As Double, ByRef sd() As Double, _
     ByRef rss As Double, ByRef objf As Double, ByRef df As Double, _
     ByVal nfv As Integer, ByRef fva() As Double, ByRef fsd() As Double, _
     ByRef rmsy As Double, ByRef options As Nag_G13_Opt)
        Dim tdxxy As Integer, nxxy As Integer, ly As Integer, nseries As Integer
        Dim npara As Integer
        Dim transfv As Nag_TransfOrder

        ' Initialize an error structure
        Dim fail As New NagError

        ' NPARA is the length of PARA
        npara = para.GetLength(0)

        ly = y.GetLength(0)

        ' Length of time series is assumed to be the number of rows in Y and we assume
        ' there are no input series (so only the output series)
        nxxy = y.GetLength(0)
        nseries = 1
        tdxxy = 1

        ' Reallocate memory for output arrays if required
        If (sd.GetLength(0) < npara) Then
            ReDim sd(npara - 1)
        End If

        ' Initialise the transfer structure
        Call g13byc(nseries, transfv, fail)
        If (fail.code <> x04ncc("NE_NOERROR")) Then
            Throw New NAGException("G13BYC", fail)
        End If

        ' Call the NAG library routine to fit the ARIMA model. 
        ' Need to pass the first element of any arrays
        Call g13bec(arimav, nseries, transfv, para, npara, nxxy, y, _
         tdxxy, sd, rss, objf, df, options, fail)
        If (fail.code <> x04ncc("NE_NOERROR")) Then
            Throw New NAGException("G13BEC", fail)
        End If

        ' If required perform the forecasts
        If (nfv > 0) Then
            Dim tdmrx As Integer, ldparx As Integer, tdparx As Integer
            Dim nev As Integer

            ' Reallocate memory for forecast information if required
            If (fva.GetLength(0) < nfv) Then
                ReDim fva(nfv - 1)
            End If
            If (fsd.GetLength(0) < nfv) Then
                ReDim fsd(nfv - 1)
            End If

            ' NEV is the number of initial values in series, as this is a univariate
            ' time series and we do not need to supply additional Y values, this is
            ' the same as the number of rows in XXY
            nev = nxxy

            ' As we do not have any input series, then RMSXY, MRX, PARX need not
            ' be initialise and TDMRX, LDPARX and TDPARX can be set to 1
            Dim mrx(0) As Integer, parx(0) As Double, rmsxy(0) As Double
            tdmrx = 1
            ldparx = 1
            tdparx = 1
            Call g13bjc(arimav, nseries, transfv, para, npara, _
             nev, nfv, y, tdxxy, rmsxy, mrx, tdmrx, parx, ldparx, _
             tdparx, fva, fsd, options, fail)
            If (fail.code <> x04ncc("NE_NOERROR")) Then
                Throw New NAGException("G13BJC", fail)
            End If

            ' Save the residual mean square for output series
            rmsy = rmsxy(0)
        End If
    End Sub

    ' Subroutine for a model estimation of a multi-input series
    ' No forecasting and default options
    Public Overloads Sub MultiInput( _
     ByVal arimav As Nag_ArimaOrder, _
     ByRef mt(,) As Integer, _
     ByRef para() As Double, _
     ByVal xxy(,) As Double, _
     ByRef sd() As Double, _
     ByRef rss As Double, _
     ByRef objf As Double, _
     ByRef df As Double _
    )
        Dim fxx(0, 0) As Double, rmsxy(0) As Double, parx(0, 0) As Double
        Dim fva(0) As Double, fsd(0) As Double
        Dim mrx(0, 0) As Integer

        ' Initialise the options structure
        Dim options As New Nag_G13_Opt()
        Call g13bxc(options)

        ' Turn off all printing etc
        options.list = Nag_Boolean.Nag_FALSE
        options.print_level = Nag_PrintType.Nag_NoPrint

        Call MultiInput(arimav, mt, para, xxy, sd, rss, objf, df, _
         fxx, rmsxy, mrx, parx, fva, fsd, options, False)

        ' Free the options structure. This is required as G13BE and G13BJ may allocate memory 
        ' to various elements in the structure.
        Call g13xzc(options)
    End Sub

    ' Subroutine for a model estimation of a multi-input series
    ' No forecasting and options supplied explicitly
    Public Overloads Sub MultiInput(ByVal arimav As Nag_ArimaOrder, _
     ByRef mt(,) As Integer, ByRef para() As Double, ByVal xxy(,) As Double, _
     ByRef sd() As Double, ByRef rss As Double, ByRef objf As Double, _
     ByRef df As Double, ByRef options As Nag_G13_Opt)
        Dim fxx(0, 0) As Double, rmsxy(0) As Double, parx(0, 0) As Double
        Dim fva(0) As Double, fsd(0) As Double
        Dim mrx(0, 0) As Integer

        Call MultiInput(arimav, mt, para, xxy, sd, rss, objf, df, _
        fxx, rmsxy, mrx, parx, fva, fsd, options, False)
    End Sub

    ' Subroutine for a model estimation of a multi-input series
    ' With forecasting and default options
    Public Overloads Sub MultiInput( _
     ByVal arimav As Nag_ArimaOrder, _
     ByRef mt(,) As Integer, _
     ByRef para() As Double, _
     ByVal xxy(,) As Double, _
     ByRef sd() As Double, _
     ByRef rss As Double, _
     ByRef objf As Double, _
     ByRef df As Double, _
     ByVal fxx(,) As Double, _
     ByRef rmsxy() As Double, _
     ByVal mrx(,) As Integer, _
     ByVal parx(,) As Double, _
     ByRef fva() As Double, _
     ByRef fsd() As Double, _
     Optional ByVal ForeCast As Boolean = True _
    )
        ' Initialise the options structure
        Dim options As New Nag_G13_Opt()
        Call g13bxc(options)

        ' Turn off all printing etc
        options.list = Nag_Boolean.Nag_FALSE
        options.print_level = Nag_PrintType.Nag_NoPrint

        Call MultiInput(arimav, mt, para, xxy, sd, rss, objf, df, _
        fxx, rmsxy, mrx, parx, fva, fsd, options, False)

        ' Free the options structure. This is required as G13BE and G13BJ may allocate memory 
        ' to various elements in the structure.
        Call g13xzc(options)
    End Sub

    ' Subroutine for a model estimation of a multi-input series
    ' With forecasting and options supplied explicitly
    Public Overloads Sub MultiInput( _
     ByVal arimav As Nag_ArimaOrder, _
     ByRef mt(,) As Integer, _
     ByRef para() As Double, _
     ByVal xxy(,) As Double, _
     ByRef sd() As Double, _
     ByRef rss As Double, _
     ByRef objf As Double, _
     ByRef df As Double, _
     ByVal fxx(,) As Double, _
     ByRef rmsxy() As Double, _
     ByVal mrx(,) As Integer, _
     ByVal parx(,) As Double, _
     ByRef fva() As Double, _
     ByRef fsd() As Double, _
     ByRef options As Nag_G13_Opt, _
     Optional ByVal ForeCast As Boolean = True _
    )
        Dim tdxxy As Integer, nxxy As Integer, nseries As Integer, npara As Integer
        Dim k As Integer, i As Integer, j As Integer, nfv As Integer
        Dim lt_xxy As Integer
        Dim t_xxy() As Double
        Dim transfv As Nag_TransfOrder

        ' Initialize an error structure
        Dim fail As New NagError

        ' NPARA is the length of PARA
        npara = para.GetLength(0)

        ' Length of time series is assumed to be the number of rows in XXY and
        ' the number of series as the number of columns in XXY
        nxxy = xxy.GetLength(0)
        nseries = xxy.GetLength(1)

        ' Number of values to forecast
        If (ForeCast) Then
            nfv = fxx.GetLength(0)
        Else
            nfv = 0
        End If

        ' Overall length required in the T_XXY matrix (G13BJ requires the observed
        ' information supplied in XXY and the future information supplied in FXX
        ' to be in the same matrix)
        lt_xxy = nxxy + nfv

        ' XXY is a two dimensional array so copy XXY into a one dimensional array
        ' making sure that the data in T_XXY is stored in row major order
        ReDim t_xxy(nseries * lt_xxy - 1)

        ' Copy the observed information
        k = -1
        For i = 0 To nxxy - 1
            For j = 0 To nseries - 1
                k = k + 1
                t_xxy(k) = xxy(i, j)
            Next j
        Next i

        ' Data is copied into T_XXY with no unused elements, so stride is the number
        ' of series
        tdxxy = nseries

        ' Reallocate memory for output arrays, if required
        If (sd.GetLength(0) < npara) Then
            ReDim sd(npara - 1)
        End If

        ' Initialise the transfer structure
        Call g13byc(nseries, transfv, fail)
        If (fail.code <> x04ncc("NE_NOERROR")) Then
            Throw New NAGException("G13BYC", fail)
        End If

        ' Copy the information from the input array MT into these arrays
        ' MT is assumed to have the same structure as the argument with the same
        ' name in the NAG Fortran routine G13BEF. This allows the data to be moved
        ' in one go without having to worry whether MT is in row or column major order
        Dim p() As Integer, q() As Integer, r() As Integer, b() As Integer
        ReDim p(nseries - 2), q(nseries - 2), r(nseries - 2), b(nseries - 2) '
        For i = 0 To nseries - 2
            b(i) = mt(0, i)
            p(i) = mt(1, i)
            q(i) = mt(2, i)
            r(i) = mt(3, i)
        Next

        ' Copy the information from MT into TRANSFV
        Call Marshal.Copy(b, 0, transfv.b, nseries - 1)
        Call Marshal.Copy(p, 0, transfv.p, nseries - 1)
        Call Marshal.Copy(q, 0, transfv.q, nseries - 1)
        Call Marshal.Copy(r, 0, transfv.r, nseries - 1)

        Dim total As Double
        total = 0
        For i = 0 To t_xxy.GetLength(0) - 1
            total = total + t_xxy(i)
        Next

        ' Call the NAG library routine
        Call g13bec(arimav, nseries, transfv, para, npara, nxxy, t_xxy, _
         tdxxy, sd, rss, objf, df, options, fail)

        ' If an error occurred in NAG routine, then throw an exception
        If (fail.code <> x04ncc("NE_NOERROR")) Then
            Throw New NAGException("G13BEC", fail)
        End If

        ' If required perform the forecasts
        If (ForeCast) Then
            Dim tdmrx As Integer, ldparx As Integer, tdparx As Integer
            Dim nev As Integer, ncol As Integer
            Dim t_mrx() As Integer, t_parx() As Double

            ' Make sure we don't access past end of array when copying FXX
            ncol = Math.Min(nseries, fxx.GetLength(1))

            ' Copy the future information into T_XXY
            k = nxxy * nseries - 1
            For i = 0 To nfv - 1
                For j = 0 To ncol - 1
                    k = k + 1
                    t_xxy(k) = fxx(i, j)
                Next j
                ' Skip one value that corresponds to the input series (Y)
                k = k + 1
            Next i

            ' As we are copying the data into arrays, and using all elements we 
            ' can set the strides to the minimum values
            tdmrx = nseries - 1
            tdparx = nseries - 1

            ' Set LDPARX to the number of rows in PARX
            ldparx = parx.GetLength(0)

            ReDim t_mrx(7 * tdmrx - 1), t_parx(tdparx * ldparx - 1)

            ' Need to copy the two dimensional input arrays MRX and PARX into one
            ' dimensional arrays in row major order
            ncol = Math.Min(tdmrx, mrx.GetLength(1))
            k = -1
            For i = 0 To 6
                For j = 0 To ncol - 1
                    k = k + 1
                    t_mrx(k) = mrx(i, j)
                Next j
            Next i
            ncol = Math.Min(tdparx, parx.GetLength(1))
            k = -1
            For i = 0 To ldparx - 1
                For j = 0 To ncol - 1
                    k = k + 1
                    t_parx(k) = parx(i, j)
                Next
            Next

            ' Reallocate memory for forecast information if required
            If (fva.GetLength(0) < nfv) Then
                ReDim fva(nfv - 1)
            End If
            If (fsd.GetLength(0) < nfv) Then
                ReDim fsd(nfv - 1)
            End If

            ' NEV is the number of initial values in series
            nev = nxxy

            ' Call the forecasting routine
            Call g13bjc(arimav, nseries, transfv, para, npara, _
            nev, nfv, t_xxy, tdxxy, rmsxy, t_mrx, tdmrx, t_parx, ldparx, _
            tdparx, fva, fsd, options, fail)
            If (fail.code <> x04ncc("NE_NOERROR")) Then
                Throw New NAGException("G13BJC", fail)
            End If
        End If

    End Sub
End Class
